type Condition @entity {
  "conditionId"
  id: ID!
}

# Neg Risk Events/Markets
type NegRiskEvent @entity {
  "negRiskMarketId"
  id: ID!
  "Fee Bps"
  feeBps: BigInt!
  "Question Count"
  questionCount: Int!
}

# Market Open Interest
type MarketOpenInterest @entity {
  "Condition ID"
  id: ID!
  "Open interest for the market"
  amount: BigInt!
}

# Global Open Interest
type GlobalOpenInterest @entity {
  "Empty string"
  id: ID!
  "Global Open Interest"
  amount: BigInt!
}