CUSTOM_ERROR_DICT = {
    "0x3a81d6fc": "AlreadyRegistered()",
    "0xcd4e6167": "FeeTooHigh()",
    "0x66f8620a": "InvalidComplement()",
    "0x756688fe": "InvalidNonce()",
    "0x8baa579f": "InvalidSignature()",
    "0x3f6cc768": "InvalidTokenId()",
    "0xe2cc6ad6": "MakingGtRemaining()",
    "0xa0b94465": "MismatchedTokenIds()",
    "0x7bfa4b9f": "NotAdmin()",
    "0x7f9a6f46": "NotCrossing()",
    "0x7c214f04": "NotOperator()",
    "0x30cd7471": "NotOwner()",
    "0x5211a079": "NotTaker()",
    "0xc56873ba": "OrderExpired()",
    "0x7b38b76e": "OrderFilledOrCancelled()",
    "0x9e87fac8": "Paused()",
    "0xdf4d8080": "TooLittleTokensReceived()",
    "0xb36eb036": "DeterminedFlagAlreadySet()",
    "0xfc520af5": "FeeBipsOutOfBounds()",
    "0x4e23d035": "IndexOutOfBounds()",
    "0x9667d381": "InvalidIndexSet()",
    "0xff633a38": "LengthMismatch()",
    "0xe3b02389": "MarketAlreadyDetermined()",
    "0xd2b4c0c6": "MarketAlreadyPrepared()",
    "0xb6649495": "MarketNotPrepared()",
    "0x24f2dfe7": "NoConvertiblePositions()",
    "0xcd7769ff": "NotApprovedForAll()",
    "0x80fee105": "OnlyOracle()",
    "0x155a4b5c": "UnexpectedCollateralToken()",
    }
